# Unit test predict_bootstrapping ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression
from lightgbm import LGBMRegressor

from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.exceptions import UnknownLevelWarning
from skforecast.preprocessing import RollingFeatures
from ....recursive import ForecasterRecursiveMultiSeries

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_wide_dt,
    series_dict_range,
    exog_wide_range,
    exog_pred_wide_range,
    expected_df_to_long_format
)

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


def test_predict_bootstrapping_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, levels=None)


def test_predict_bootstrapping_IgnoredArgumentWarning_when_not_available_self_last_window_for_some_levels():
    """
    Test IgnoredArgumentWarning is raised when last_window is not available for 
    levels because it was not stored during fit.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], 
        store_last_window=['l1'], store_in_sample_residuals=True
    )

    warn_msg = re.escape(
        "Levels {'l2'} are excluded from prediction "
        "since they were not stored in `last_window_` attribute "
        "during training. If you don't want to retrain the "
        "Forecaster, provide `last_window` as argument."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        results = forecaster.predict_bootstrapping(
                      steps                   = 1,
                      levels                  = ['l1', 'l2'],
                      n_boot                  = 4, 
                      exog                    = exog_pred_wide_range['exog_1'], 
                      use_in_sample_residuals = True,
                      use_binned_residuals    = False
                  )

    expected = {
        'l1': pd.DataFrame(
                  data    = np.array([[0.48128438, 0.20172898, 0.07381993, 0.38819631]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(start=50, stop=51)
              )
    }
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("store_last_window",
                         [['1'], False],
                         ids=lambda slw: f"store_last_window: {slw}")
def test_predict_bootstrapping_ValueError_when_not_available_self_last_window_for_levels(store_last_window):
    """
    Test ValueError is raised when last_window is not available for all 
    levels because it was not stored during fit.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=5)
    forecaster.fit(series=series_dict_range, store_last_window=store_last_window)

    err_msg = re.escape(
        "No series to predict. None of the series {'l2'} are present in "
        "`last_window_` attribute. Provide `last_window` as argument "
        "in predict method."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=5, levels=['l2'], last_window=None)


def test_predict_bootstrapping_IgnoredArgumentWarning_when_levels_is_list_and_different_last_index_in_self_last_window_DatetimeIndex():
    """
    Test IgnoredArgumentWarning is raised when levels is a list and have 
    different last index in last_window attribute using a DatetimeIndex.
    """
    series_2 = {
        '1': series_wide_dt['1'].copy(),
        '2': series_wide_dt['2'].iloc[:30].copy()
    }
    series_2['1'].index = pd.date_range(start='2020-01-01', periods=50)
    series_2['2'].index = pd.date_range(start='2020-01-01', periods=30)
    exog_2 = {
        '1': exog_wide_range['exog_1'].copy(),
        '2': exog_wide_range['exog_1'].iloc[:30].copy()
    }
    exog_2['1'].index = pd.date_range(start='2020-01-01', periods=50)
    exog_2['2'].index = pd.date_range(start='2020-01-01', periods=30)
    exog_2_pred = {
        '1': exog_pred_wide_range['exog_1'].copy()
    }
    exog_2_pred['1'].index = pd.date_range(start='2020-02-20', periods=50)

    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(series=series_2, exog=exog_2, store_in_sample_residuals=True)

    warn_msg = re.escape(
        "Only series whose last window ends at the same index "
        "can be predicted together. Series that do not reach the "
        "maximum index, '2020-02-19 00:00:00', are excluded "
        "from prediction: {'2'}."
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        results = forecaster.predict_bootstrapping(
                      steps                   = 1,
                      levels                  = ['1', '2'],
                      n_boot                  = 4, 
                      exog                    = exog_2_pred, 
                      use_in_sample_residuals = True,
                      use_binned_residuals    = False
                  )

    expected = {
        '1': pd.DataFrame(
                 data    = np.array([[0.47931483, 0.21880114, 0.10692630, 0.42510201]]),
                 columns = [f"pred_boot_{i}" for i in range(4)],
                 index   = pd.date_range(start='2020-02-20', periods=1, freq=None)
             )
    }
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_UnknownLevelWarning_when_not_in_sample_residuals_for_level():
    """
    Test UnknownLevelWarning is raised when predicting an unknown level and 
    use_in_sample_residuals is True (using _unknown_level residuals).
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LGBMRegressor(verbose=-1), lags=3, encoding='ordinal'
    )
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)
    last_window = pd.DataFrame(forecaster.last_window_)
    last_window['l3'] = last_window['l1']

    warn_msg = re.escape(
        "`levels` {'l3'} were not included in training. "
        "Unknown levels are encoded as NaN, which may cause the "
        "prediction to fail if the estimator does not accept NaN values."
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        results = forecaster.predict_bootstrapping(
                      steps                   = 1,
                      levels                  = ['l1', 'l2', 'l3'],
                      last_window             = last_window,
                      n_boot                  = 4,
                      use_in_sample_residuals = True,
                      use_binned_residuals    = False
                  )

    warn_msg = re.escape(
        "`levels` {'l3'} are not present in `forecaster.in_sample_residuals_`, "
        "most likely because they were not present in the training data. "
        "A random sample of the residuals from other levels will be used. "
        "This can lead to inaccurate intervals for the unknown levels."
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        results = forecaster.predict_bootstrapping(
                      steps                   = 1,
                      levels                  = ['l1', 'l2', 'l3'],
                      last_window             = last_window,
                      n_boot                  = 4,
                      use_in_sample_residuals = True,
                      use_binned_residuals    = False
                  )

    expected = {
        'l1': pd.DataFrame(
                  data    = np.array([[0.60487599, 0.22612695, 0.07322577, 0.45340699]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(50, 51)
              ),
        'l2': pd.DataFrame(
                  data    = np.array([[0.69343437, 0.67766497, 0.69444735, 0.64415095]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(50, 51)
              ),
        'l3': pd.DataFrame(
                  data    = np.array([[0.26291992, 0.45547833, 0.22612695, 0.51589959]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(50, 51)
              )
    }
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_encoding_None_unknown_level():
    """
    Test UnknownLevelWarning is raised when predicting an unknown level and 
    encoding is None (using _unknown_level residuals).
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LGBMRegressor(verbose=-1), lags=3, encoding=None
    )
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)
    last_window = pd.DataFrame(forecaster.last_window_)
    last_window['l3'] = last_window['l1']

    results = forecaster.predict_bootstrapping(
                  steps                   = 1,
                  levels                  = ['l1', 'l2', 'l3'],
                  last_window             = last_window,
                  n_boot                  = 4,
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )

    expected = {
        'l1': pd.DataFrame(
                  data    = np.array([[0.62339220, 0.03921391, 0.58593655, 0.45059094]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(50, 51)
              ),
        'l2': pd.DataFrame(
                  data    = np.array([[0.94743870, 0.53474394, 0.38578090, 0.56857677]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(50, 51)
              ),
        'l3': pd.DataFrame(
                  data    = np.array([[0.23759071, 0.42972248, 0.19361364, 0.49593259]]),
                  columns = [f"pred_boot_{i}" for i in range(4)],
                  index   = pd.RangeIndex(50, 51)
              )
    }
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_encoding_None_unknown_level_differentiation():
    """
    Test UnknownLevelWarning is raised when predicting an unknown level and 
    encoding is None (using _unknown_level residuals) and differentiation=1.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LGBMRegressor(verbose=-1), lags=3, encoding=None, differentiation=1
    )
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)
    last_window = pd.DataFrame(forecaster.last_window_)
    last_window['l3'] = last_window['l1'] * 0.9

    results = forecaster.predict_bootstrapping(
                  steps                   = 1,
                  levels                  = ['l1', 'l2', 'l3'],
                  last_window             = last_window,
                  n_boot                  = 5,
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )

    expected = {
        'l1': pd.DataFrame(
                  data    = np.array([[0.63878989, 0.33256144, 0.62392879, 0.37577094, 1.14849867]]),
                  columns = [f"pred_boot_{i}" for i in range(5)],
                  index   = pd.RangeIndex(50, 51)
              ),
        'l2': pd.DataFrame(
                  data    = np.array([[0.26274688, -0.00150092, 0.26392699, 0.18584762, 0.26392699]]),
                  columns = [f"pred_boot_{i}" for i in range(5)],
                  index   = pd.RangeIndex(50, 51)
              ),
        'l3': pd.DataFrame(
                  data    = np.array([[0.44105054, 0.54865001, 0.26752194, 0.36550593, 0.26752194]]),
                  columns = [f"pred_boot_{i}" for i in range(5)],
                  index   = pd.RangeIndex(50, 51)
              )
    }
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_ValueError_when_out_sample_residuals_is_None():
    """
    Test ValueError is raised when use_in_sample_residuals=False and
    forecaster.out_sample_residuals_ is None.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)

    err_msg = re.escape(
        "`forecaster.out_sample_residuals_` is either None or empty. Use "
        "`use_in_sample_residuals = True` or the "
        "`set_out_sample_residuals()` method before predicting."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(
            steps=1, levels='l1', use_in_sample_residuals=False, use_binned_residuals=False
        )


def test_predict_bootstrapping_UnknownLevelWarning_out_sample_residuals_with_encoding_None():
    """
    Test UnknownLevelWarning is raised when encoding is None and 
    out_sample_residuals is set.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=None
    )
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)

    y_true = {
        'l1': np.array([1, 2, 3, 4, 5]), 
        'l2': np.array([1, 2, 3, 4, 5])
    }
    y_pred = {
        'l1': np.array([0, 0, 0, 0, 0]),
        'l2': np.array([0, 0, 0, 0, 0])
    }
    warn_msg = re.escape(
        "As `encoding` is set to `None`, no distinction between levels "
        "is made. All residuals are stored in the '_unknown_level' key."
    )
    with pytest.warns(UnknownLevelWarning, match = warn_msg):
        forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)
    
    last_window = pd.DataFrame(forecaster.last_window_)
    last_window['l3'] = last_window['l1']
    results = forecaster.predict_bootstrapping(
                  steps                   = 1,
                  levels                  = ['l1', 'l2', 'l3'],
                  last_window             = last_window,
                  n_boot                  = 4,
                  use_in_sample_residuals = False,
                  use_binned_residuals    = False
              )

    expected = {
        'l1': pd.DataFrame(
               data    = np.array([[1.48307367, 2.48307367, 1.48307367, 1.48307367]]),
               columns = [f"pred_boot_{i}" for i in range(4)],
               index   = pd.RangeIndex(50, 51)
           ),
        'l2': pd.DataFrame(
               data    = np.array([[5.5061193, 3.5061193, 3.5061193, 2.5061193]]),
               columns = [f"pred_boot_{i}" for i in range(4)],
               index   = pd.RangeIndex(50, 51)
           ),
        'l3': pd.DataFrame(
               data    = np.array([[4.48307367, 2.48307367, 4.48307367, 4.48307367]]),
               columns = [f"pred_boot_{i}" for i in range(4)],
               index   = pd.RangeIndex(50, 51)
           )
    }
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("transformer_series", 
                         [None, StandardScaler()],
                         ids = lambda tr: f'transformer_series type: {type(tr)}')
def test_predict_bootstrapping_ValueError_when_not_level_in_out_sample_residuals(transformer_series):
    """
    Test ValueError is raised when use_in_sample_residuals=False and
    forecaster.out_sample_residuals_ is missing a level.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=transformer_series
    )
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)
    y_true = {'l1': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([0, 0, 0, 0, 0])}
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)

    err_msg = re.escape(
        "Residuals for level 'l2' are None. Check `forecaster.out_sample_residuals_`."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(
            steps=3, use_in_sample_residuals=False, use_binned_residuals=False
        )


@pytest.mark.parametrize("transformer_series", 
                         [None, StandardScaler()],
                         ids = lambda tr: f'transformer_series type: {type(tr)}')
def test_predict_bootstrapping_ValueError_when_not_level_in_out_sample_residuals_binned_true(transformer_series):
    """
    Test ValueError is raised when use_in_sample_residuals=False and
    forecaster.out_sample_residuals_ is missing a level.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(),
        lags=3,
        transformer_series=transformer_series
    )
    forecaster.fit(series=series_dict_range, store_in_sample_residuals=True)
    y_true = {'l1': np.array([1, 2, 3, 4, 5])}
    y_pred = {'l1': np.array([0, 0, 0, 0, 0])}
    forecaster.set_out_sample_residuals(y_true=y_true, y_pred=y_pred)

    err_msg = re.escape(
        "Residuals for level 'l2' are None. Check `forecaster.out_sample_residuals_by_bin_`."
    )

    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=3, use_in_sample_residuals=False, use_binned_residuals=True)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_True():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    1 step ahead is predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 1, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range['exog_1'], 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.48128438, 0.20172898, 0.07381993, 0.38819631]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.72210743, 0.73878038, 0.80513315, 0.57237401]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_True_binned_True():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    1 step ahead is predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 1, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range['exog_1'], 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = True
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.38187547242082853, 0.4143065346074915, 0.4289575462130275, 0.38187547242082853]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.6804435739774055, 0.07849209244403665, 0.6804435739774055, 0.07849209244403665]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_True():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    2 steps ahead are predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 2, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range['exog_1'], 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.48128438, 0.20172898, 0.07381993, 0.38819631],
                                         [0.22472696, 0.06269684, 0.35120532, 0.66839553]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.08624916, 0.57237401, 0.69862572, 0.92542401],
                                         [0.06877403, 0.19830837, 0.10900956, 0.09861192]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )    

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_True_binned_True():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    2 steps ahead are predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                    steps                   = 2, 
                    n_boot                  = 4, 
                    exog                    = exog_pred_wide_range['exog_1'], 
                    use_in_sample_residuals = True,
                    use_binned_residuals    = True
                )

    expected_1 = pd.DataFrame(
                        data    = np.array([[0.1638659133321268, 0.37782728733797316, 0.38187547242082853, 0.37782728733797316],
                                            [0.06020908790902979, 0.4944130549880117, 0.4925928380713404, 0.22243700400180003]]),
                        columns = [f"pred_boot_{i}" for i in range(4)],
                        index   = pd.RangeIndex(start=50, stop=52)
                    )
    expected_2 = pd.DataFrame(
                        data    = np.array([[0.3180664981712658, 0.6804435739774055, 0.6966346511864026, 0.5723740055300023],
                                            [0.5805477599371661, 0.68446723113257, 0.7017221250999669, 0.3992856989330053]]),
                        columns = [f"pred_boot_{i}" for i in range(4)],
                        index   = pd.RangeIndex(start=50, stop=52)
                    )    

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')
    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_False():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    1 step ahead is predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )    
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    results = forecaster.predict_bootstrapping(
                  steps                   = 1, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range['exog_1'], 
                  use_in_sample_residuals = False,
                  use_binned_residuals    = False,
                  suppress_warnings       = True
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.48128438, 0.20172898, 0.07381993, 0.38819631]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.72210743, 0.73878038, 0.80513315, 0.57237401]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_False():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    2 steps ahead are predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    forecaster.out_sample_residuals_ = forecaster.in_sample_residuals_
    results = forecaster.predict_bootstrapping(
                  steps                   = 2, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range['exog_1'], 
                  use_in_sample_residuals = False,
                  use_binned_residuals    = False
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([[0.48128438, 0.20172898, 0.07381993, 0.38819631],
                                         [0.22472696, 0.06269684, 0.35120532, 0.66839553]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.08624916, 0.57237401, 0.69862572, 0.92542401],
                                         [0.06877403, 0.19830837, 0.10900956, 0.09861192]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )    

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_False_binned_True():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression and
    2 steps ahead are predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, transformer_series=None
    )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_
    results = forecaster.predict_bootstrapping(
                    steps                   = 2, 
                    n_boot                  = 4, 
                    exog                    = exog_pred_wide_range['exog_1'], 
                    use_in_sample_residuals = False,
                    use_binned_residuals    = True,
                )

    expected_1 = pd.DataFrame(
                        data    = np.array([[0.1638659133321268, 0.37782728733797316, 0.38187547242082853, 0.37782728733797316],
                                            [0.06020908790902979, 0.4944130549880117, 0.4925928380713404, 0.22243700400180003]]),
                        columns = [f"pred_boot_{i}" for i in range(4)],
                        index   = pd.RangeIndex(start=50, stop=52)
                    )
    expected_2 = pd.DataFrame(
                        data    = np.array([[0.3180664981712658, 0.6804435739774055, 0.6966346511864026, 0.5723740055300023],
                                            [0.5805477599371661, 0.68446723113257, 0.7017221250999669, 0.3992856989330053]]),
                        columns = [f"pred_boot_{i}" for i in range(4)],
                        index   = pd.RangeIndex(start=50, stop=52)
                    )    

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer_fixed_to_zero():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included, both
    inputs are transformed and in-sample residuals are fixed to 0.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )

    forecaster.in_sample_residuals_['l1'] = np.full_like(
        forecaster.in_sample_residuals_['l1'], fill_value=0
    )
    forecaster.in_sample_residuals_['l2'] = np.full_like(
        forecaster.in_sample_residuals_['l2'], fill_value=0
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 2, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([[0.50201669, 0.50201669, 0.50201669, 0.50201669],
                                         [0.49804821, 0.49804821, 0.49804821, 0.49804821]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.52531076, 0.52531076, 0.52531076, 0.52531076],
                                         [0.51683613, 0.51683613, 0.51683613, 0.51683613]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer():
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 2, 
                  n_boot                  = 4, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([[0.48598851, 0.24503818, 0.10778738, 0.38990777],
                                         [0.25740769, 0.06499781, 0.35491959, 0.67049609]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.08931035, 0.57081569, 0.70066373, 0.96406103],
                                         [0.06876068, 0.23752454, 0.11044965, 0.10202009]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_window_features():
    """
    Test output of predict_bootstrapping when estimator is LGBMRegressor 
    and window features.
    """

    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=4)
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(verbose=-1),
                     lags               = 5,
                     window_features    = rolling,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 6, 
                  n_boot                  = 10, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )

    expected_1 = pd.DataFrame(
                     data = np.array([
                                [0.70414275, 0.33846487, 0.51525768, 0.78238922, 0.53358641,
                                 0.55496755, 0.49698665, 0.3156552 , 0.70164044, 0.51987002],
                                [0.5093462 , 0.10945852, 0.50986314, 0.53323641, 0.50986314,
                                 0.31517078, 0.54867791, 0.09113636, 0.46379211, 0.41342218],
                                [0.86504059, 0.56932758, 0.45218404, 0.44580085, 0.83119047,
                                 0.43019077, 0.68569118, 0.51340363, 0.61452721, 0.84957527],
                                [0.65524631, 0.84194486, 0.93082861, 0.7155279 , 0.83658981,
                                 1.02306043, 0.6582736 , 0.86211073, 0.56189833, 0.70564327],
                                [0.54647555, 0.75265845, 0.43334796, 0.36702491, 0.56762363,
                                 0.43184972, 0.66142918, 0.47738529, 0.57800408, 0.17483813],
                                [0.53343541, 0.56488173, 0.78929977, 0.44786181, 0.49626976,
                                 0.6850046 , 0.36524591, 0.70909355, 0.34761764, 0.39494035]]),
                     index   = pd.RangeIndex(start=50, stop=56, step=1),
                     columns = [f"pred_boot_{i}" for i in range(10)]
                 )
    expected_2 = pd.DataFrame(
                     data = np.array([
                                [0.4703539 , 0.61121897, 0.7456648 , 1.05767197, 0.85702976,
                                 0.84549833, 0.90211559, 1.05767197, 0.84549833, 0.43576797],
                                [0.17638536, 0.36323208, 0.28360047, 0.39694573, 0.42021445,
                                 0.62085666, 0.21910573, 0.1686658 , 0.03353859, 0.17638536],
                                [0.63706768, 0.95491448, 0.84158046, 0.88270255, 0.63313527,
                                 0.56008641, 0.4801817 , 0.64017749, 0.48757368, 0.70482518],
                                [0.88520648, 0.81333217, 0.27723066, 0.44112079, 0.45918293,
                                 0.57023604, 0.96036122, 0.63876782, 0.81917714, 0.82440159],
                                [0.45118834, 0.55605023, 0.84244033, 0.9414499 , 0.55014026,
                                 0.50114166, 0.25916732, 0.52348578, 0.64211623, 0.49025447],
                                [0.49198322, 0.10785838, 0.12577904, 0.12725819, 0.79780591,
                                 0.54327887, 0.78210044, 0.78482102, 0.19724282, 0.7278225 ]]),
                     index   = pd.RangeIndex(start=50, stop=56, step=1),
                     columns = [f"pred_boot_{i}" for i in range(10)]
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_window_features_in_sample_residuals_True_binned_True():
    """
    Test output of predict_bootstrapping when estimator is LGBMRegressor 
    and window features.
    """

    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=4)
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(verbose=-1),
                     lags               = 5,
                     window_features    = rolling,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 6, 
                  n_boot                  = 10, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = True
              )

    expected_1 = pd.DataFrame(
                     data = np.array([
                                [0.61785158, 0.92107198, 0.33846487, 0.61785158, 0.61785158,
                                0.61785158, 0.33846487, 0.71098463, 0.33846487, 0.61785158],
                                [0.40203929, 0.51815394, 0.34438401, 0.39861071, 0.39861071,
                                0.40203929, 0.51861896, 0.51815394, 0.35853838, 0.39861071],
                                [0.60985531, 0.6193728 , 0.73992069, 0.69381973, 0.69381973,
                                0.42961981, 0.37331125, 0.69849163, 0.75557212, 0.69381973],
                                [0.91012366, 0.45240191, 0.91650657, 0.89227671, 0.97437054,
                                0.96084168, 0.8955076 , 0.5144838 , 0.84243625, 0.86350256],
                                [0.40947858, 0.53993477, 0.39449316, 0.41415048, 0.4107219 ,
                                0.31623829, 0.62420462, 0.42565511, 0.49978962, 0.41415048],
                                [0.39102081, 0.41718334, 0.5183308 , 0.48717554, 0.41773124,
                                0.33614795, 0.70731688, 0.33183784, 0.73521587, 0.16788473]]),
                     index   = pd.RangeIndex(start=50, stop=56, step=1),
                     columns = [f"pred_boot_{i}" for i in range(10)]
                 )
    expected_2 = pd.DataFrame(
                     data = np.array([
                                [0.96793066,  0.85237552,  0.84549833,  0.85237552,  0.84549833,
                                0.85237552,  0.84549833,  0.72041578,  0.72041578,  0.84549833],
                                [0.03862877,  0.22038531, -0.00210693,  0.16960244, -0.00210693,
                                0.17440366,  0.17440366,  0.22038531, -0.00210693, -0.00210693],
                                [0.91961227,  0.91961227,  0.91961227,  0.70366831,  0.72180936,
                                0.54177164,  0.59375641,  0.59375641,  0.72180936,  0.72180936],
                                [0.66983588,  0.50793921,  0.88577984,  1.00028073,  0.55995266,
                                0.98041187,  0.66322437,  0.83652287,  0.87764896,  0.88580852],
                                [0.53785703,  0.66320532,  0.62033826,  0.56929985,  0.86432032,
                                0.88789079,  0.70306078,  0.55964353,  0.64136831,  0.62033826],
                                [0.69776045,  0.43383701,  0.30191143,  0.30191143,  0.00903454,
                                0.68260119,  0.10580831,  0.37015406,  0.67620055,  0.24612479]]),
                     index   = pd.RangeIndex(start=50, stop=56, step=1),
                     columns = [f"pred_boot_{i}" for i in range(10)]
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


def test_predict_bootstrapping_output_when_window_features_in_sample_residuals_False_binned_True():
    """
    Test output of predict_bootstrapping when estimator is LGBMRegressor 
    and window features.
    """

    rolling = RollingFeatures(stats=['mean', 'median'], window_sizes=4)
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LGBMRegressor(verbose=-1),
                     lags               = 5,
                     window_features    = rolling,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    forecaster.out_sample_residuals_by_bin_ = forecaster.in_sample_residuals_by_bin_
    results = forecaster.predict_bootstrapping(
                  steps                   = 6, 
                  n_boot                  = 10, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = False,
                  use_binned_residuals    = True
              )

    expected_1 = pd.DataFrame(
                     data = np.array([
                                [0.61785158, 0.92107198, 0.33846487, 0.61785158, 0.61785158,
                                0.61785158, 0.33846487, 0.71098463, 0.33846487, 0.61785158],
                                [0.40203929, 0.51815394, 0.34438401, 0.39861071, 0.39861071,
                                0.40203929, 0.51861896, 0.51815394, 0.35853838, 0.39861071],
                                [0.60985531, 0.6193728 , 0.73992069, 0.69381973, 0.69381973,
                                0.42961981, 0.37331125, 0.69849163, 0.75557212, 0.69381973],
                                [0.91012366, 0.45240191, 0.91650657, 0.89227671, 0.97437054,
                                0.96084168, 0.8955076 , 0.5144838 , 0.84243625, 0.86350256],
                                [0.40947858, 0.53993477, 0.39449316, 0.41415048, 0.4107219 ,
                                0.31623829, 0.62420462, 0.42565511, 0.49978962, 0.41415048],
                                [0.39102081, 0.41718334, 0.5183308 , 0.48717554, 0.41773124,
                                0.33614795, 0.70731688, 0.33183784, 0.73521587, 0.16788473]]),
                     index   = pd.RangeIndex(start=50, stop=56, step=1),
                     columns = [f"pred_boot_{i}" for i in range(10)]
                 )
    expected_2 = pd.DataFrame(
                     data = np.array([
                                [0.96793066,  0.85237552,  0.84549833,  0.85237552,  0.84549833,
                                0.85237552,  0.84549833,  0.72041578,  0.72041578,  0.84549833],
                                [0.03862877,  0.22038531, -0.00210693,  0.16960244, -0.00210693,
                                0.17440366,  0.17440366,  0.22038531, -0.00210693, -0.00210693],
                                [0.91961227,  0.91961227,  0.91961227,  0.70366831,  0.72180936,
                                0.54177164,  0.59375641,  0.59375641,  0.72180936,  0.72180936],
                                [0.66983588,  0.50793921,  0.88577984,  1.00028073,  0.55995266,
                                0.98041187,  0.66322437,  0.83652287,  0.87764896,  0.88580852],
                                [0.53785703,  0.66320532,  0.62033826,  0.56929985,  0.86432032,
                                0.88789079,  0.70306078,  0.55964353,  0.64136831,  0.62033826],
                                [0.69776045,  0.43383701,  0.30191143,  0.30191143,  0.00903454,
                                0.68260119,  0.10580831,  0.37015406,  0.67620055,  0.24612479]]),
                     index   = pd.RangeIndex(start=50, stop=56, step=1),
                     columns = [f"pred_boot_{i}" for i in range(10)]
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("differentiation", 
                         [1, {'l1': 1, 'l2': 1, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_predict_bootstrapping_output_when_differentiation(differentiation):
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed and differentiation.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                     differentiation    = differentiation
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 3, 
                  n_boot                  = 10, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = False
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([
                                   [0.86835554,  0.43568881,  0.72677672,  1.19824354,  0.44801019,
                                    0.66116832,  0.98730031,  0.49328573,  0.83294094,  0.49328573],
                                   [0.932948  , -0.04286105,  0.78881897,  0.94368941,  0.68990191,
                                    0.39479453,  0.66796447, -0.02242345,  0.79317039,  1.09345254],
                                   [0.80047137,  0.1669375 ,  0.49039454,  1.27844361,  1.00682521,
                                   -0.02294521,  0.66800615,  0.0497091 ,  0.79972101,  1.18391928]]),
                     columns = [f"pred_boot_{i}" for i in range(10)],
                     index   = pd.RangeIndex(start=50, stop=53)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([
                                   [0.30645781,  0.38475619, -0.21650711, -0.11175337,  0.13573138,
                                    0.45197729,  0.07709315,  0.62428333,  0.42746891, -0.02426089],
                                   [0.62617546,  0.48165268, -0.50937191,  0.02315903,  0.00976064,
                                    0.37468356, -0.01104646,  0.34859438,  0.48364255,  0.07752748],
                                   [0.2132809 ,  0.24251965, -0.01663732,  0.39527145,  0.5228262 ,
                                   -0.15531628, -0.40091162,  0.07307154,  0.35477817,  0.31644122]]),
                     columns = [f"pred_boot_{i}" for i in range(10)],
                     index   = pd.RangeIndex(start=50, stop=53)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)


@pytest.mark.parametrize("differentiation", 
                         [1, {'l1': 1, 'l2': 1, '_unknown_level': 1}], 
                         ids = lambda diff: f'differentiation: {diff}')
def test_predict_bootstrapping_output_when_differentiation_binned_residuals(differentiation):
    """
    Test output of predict_bootstrapping when estimator is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed and differentiation.
    """
    forecaster = ForecasterRecursiveMultiSeries(
                     estimator          = LinearRegression(),
                     lags               = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                     differentiation    = differentiation
                 )
    forecaster.fit(
        series=series_dict_range, exog=exog_wide_range, store_in_sample_residuals=True
    )
    results = forecaster.predict_bootstrapping(
                  steps                   = 3, 
                  n_boot                  = 10, 
                  exog                    = exog_pred_wide_range, 
                  use_in_sample_residuals = True,
                  use_binned_residuals    = True
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([
                                    [0.98730031, 1.19824354, 0.98730031, 0.98730031, 1.19824354,
                                     0.73904951, 1.19824354, 0.98730031, 0.98730031, 0.73904951],
                                    [0.50637545, 0.72305662, 1.0938956 , 0.59840128, 1.22730591,
                                     0.85164832, 0.66507416, 1.0938956 , 0.50637545, 0.92247986],
                                    [0.56426017, 0.62326123, 1.12633254, 0.64332246, 1.04217626,
                                     0.94487752, 0.91890536, 0.94378655, 0.61066816, 0.66577299]]),
                     columns = [f"pred_boot_{i}" for i in range(10)],
                     index   = pd.RangeIndex(start=50, stop=53)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([
                                    [0.43881105,  0.62428333,  0.43881105,  0.57646753,  0.57646753,
                                     0.18699591,  0.43881105,  0.57646753,  0.43881105,  0.43881105],
                                    [0.23981471,  0.26830929,  0.37001167,  0.10457883,  0.10457883,
                                     0.12916674,  0.37001167,  0.10457883,  0.02459525,  0.37001167],
                                    [0.40779716,  0.23619536,  0.45399604,  0.93833504, -0.22103177,
                                     0.2059947 ,  0.3523978 ,  0.5755924 , -0.28272742,  0.72638951]]),
                     columns = [f"pred_boot_{i}" for i in range(10)],
                     index   = pd.RangeIndex(start=50, stop=53)
                 )

    expected = {'l1': expected_1, 'l2': expected_2}
    expected = expected_df_to_long_format(expected, method='bootstrapping')

    pd.testing.assert_frame_equal(results, expected)



# Fixtures _forecaster_recursive_multiseries
# ==============================================================================
import joblib
from pathlib import Path
import numpy as np
import pandas as pd
from skforecast.preprocessing import reshape_series_wide_to_long

# Fixtures Description:
# series_wide_range: DataFrame with two time series in wide format and range index.
# series_wide_dt: DataFrame with two time series in wide format and datetime index.
# series_long_dt: MultiIndex long format DataFrame with two time series and datetime index.
# series_dict_range: Dictionary with two time series in wide format and range index.
# series_dict_dt: Dictionary with two time series in wide format and datetime index.
# exog_wide_range: DataFrame with exogenous variables in wide format and range index.
# exog_wide_dt: DataFrame with exogenous variables in wide format and datetime index.
# exog_long_dt: MultiIndex long format DataFrame with exogenous variables and datetime index.
# exog_dict_range: Dictionary with exogenous variables in wide format and range index.
# exog_dict_dt: Dictionary with exogenous variables in wide format and datetime index.
# exog_pred_wide_range: DataFrame with exogenous variables for prediction in wide format and range index.
# exog_pred_wide_dt: DataFrame with exogenous variables for prediction in wide format and datetime index.
# exog_pred_long_dt: MultiIndex long format DataFrame with exogenous variables for prediction and datetime index.
# exog_pred_dict_range: Dictionary with exogenous variables for prediction in wide format and range index.
# exog_pred_dict_dt: Dictionary with exogenous variables for prediction in wide format and datetime index.
# series_dict_nans: Dictionary with time series containing NaN values and datetime index.
# exog_dict_nans: Dictionary with exogenous variables containing NaN values and datetime index.
# series_dict_nans_train: Dictionary with training time series containing NaN values and datetime index.
# series_dict_nans_test: Dictionary with test time series containing NaN values and datetime index.
# exog_dict_nans_train: Dictionary with training exogenous variables containing NaN values and datetime index.
# exog_dict_nans_test: Dictionary with test exogenous variables containing NaN values and datetime index.

# Code to generate fixtures values:
# np.random.seed(123)
# series_wide_dt['1'] = np.random.rand(50)
# series_wide_dt['2'] = np.random.rand(50)
# exog_wide_range['exog_1'] = np.random.rand(50)
series_wide_range = pd.DataFrame(
    {'1': pd.Series(np.array(
            [0.69646919, 0.28613933, 0.22685145, 0.55131477, 0.71946897,
            0.42310646, 0.9807642 , 0.68482974, 0.4809319 , 0.39211752,
            0.34317802, 0.72904971, 0.43857224, 0.0596779 , 0.39804426,
            0.73799541, 0.18249173, 0.17545176, 0.53155137, 0.53182759,
            0.63440096, 0.84943179, 0.72445532, 0.61102351, 0.72244338,
            0.32295891, 0.36178866, 0.22826323, 0.29371405, 0.63097612,
            0.09210494, 0.43370117, 0.43086276, 0.4936851 , 0.42583029,
            0.31226122, 0.42635131, 0.89338916, 0.94416002, 0.50183668,
            0.62395295, 0.1156184 , 0.31728548, 0.41482621, 0.86630916,
            0.25045537, 0.48303426, 0.98555979, 0.51948512, 0.61289453])
        ), 
    '2': pd.Series(np.array(
            [0.12062867, 0.8263408 , 0.60306013, 0.54506801, 0.34276383,
            0.30412079, 0.41702221, 0.68130077, 0.87545684, 0.51042234,
            0.66931378, 0.58593655, 0.6249035 , 0.67468905, 0.84234244,
            0.08319499, 0.76368284, 0.24366637, 0.19422296, 0.57245696,
            0.09571252, 0.88532683, 0.62724897, 0.72341636, 0.01612921,
            0.59443188, 0.55678519, 0.15895964, 0.15307052, 0.69552953,
            0.31876643, 0.6919703 , 0.55438325, 0.38895057, 0.92513249,
            0.84167   , 0.35739757, 0.04359146, 0.30476807, 0.39818568,
            0.70495883, 0.99535848, 0.35591487, 0.76254781, 0.59317692,
            0.6917018 , 0.15112745, 0.39887629, 0.2408559 , 0.34345601])
        )
    }
)
series_wide_dt = series_wide_range.copy()
series_wide_dt.index = pd.date_range(
    start='2000-01-01', periods=len(series_wide_dt), freq='D'
)

series_dict_dt = (
    series_wide_dt
    .copy()
    .rename(columns={'1': 'l1', '2': 'l2'})
    .to_dict(orient='series')
)
series_long_dt = reshape_series_wide_to_long(series_wide_dt)
series_dict_range = {
    k: v.copy().reset_index(drop=True)
    for k, v in series_dict_dt.items()
}

exog_wide_range = pd.DataFrame(
    {'exog_1': pd.Series(np.array(
                    [0.51312815, 0.66662455, 0.10590849, 0.13089495, 0.32198061,
                    0.66156434, 0.84650623, 0.55325734, 0.85445249, 0.38483781,
                    0.3167879 , 0.35426468, 0.17108183, 0.82911263, 0.33867085,
                    0.55237008, 0.57855147, 0.52153306, 0.00268806, 0.98834542,
                    0.90534158, 0.20763586, 0.29248941, 0.52001015, 0.90191137,
                    0.98363088, 0.25754206, 0.56435904, 0.80696868, 0.39437005,
                    0.73107304, 0.16106901, 0.60069857, 0.86586446, 0.98352161,
                    0.07936579, 0.42834727, 0.20454286, 0.45063649, 0.54776357,
                    0.09332671, 0.29686078, 0.92758424, 0.56900373, 0.457412  ,
                    0.75352599, 0.74186215, 0.04857903, 0.7086974 , 0.83924335])
                ),
    'exog_2': ['a'] * 25 + ['b'] * 25}
)
exog_wide_dt = exog_wide_range.copy()
exog_wide_dt.index = pd.date_range(start='2000-01-01', periods=len(exog_wide_dt), freq='D')

exog_long_dt = exog_wide_dt.copy()
exog_long_dt.index.name = "datetime"
exog_long_dt = [exog_long_dt.assign(series_id=f"l{i}") for i in range(1, 3)]
exog_long_dt = pd.concat(exog_long_dt)
exog_long_dt = exog_long_dt.set_index(["series_id", exog_long_dt.index])

exog_dict_range = {
    'l1': exog_wide_range.copy(),
    'l2': exog_wide_range['exog_1'].copy()
}
exog_dict_dt = {
    'l1': exog_wide_dt.copy(),
    'l2': exog_wide_dt['exog_1'].copy()
}

exog_pred_wide_dt = exog_wide_dt.copy()
exog_pred_wide_dt.index = pd.date_range(
    start='2000-02-20', periods=len(exog_pred_wide_dt), freq='D'
)
exog_pred_wide_range = exog_wide_range.copy()
exog_pred_wide_range.index = pd.RangeIndex(start=50, stop=100, step=1)

exog_pred_long_dt = exog_pred_wide_dt.copy()
exog_pred_long_dt.index.name = "datetime"
exog_pred_long_dt = [exog_pred_long_dt.assign(series_id=f"l{i}") for i in range(1, 3)]
exog_pred_long_dt = pd.concat(exog_pred_long_dt)
exog_pred_long_dt = exog_pred_long_dt.set_index(["series_id", exog_pred_long_dt.index])

exog_pred_dict_range = {
    'l1': exog_pred_wide_range.copy(),
    'l2': exog_pred_wide_range['exog_1'].copy()
}
exog_pred_dict_dt = {
    'l1': exog_pred_wide_dt.copy(),
    'l2': exog_pred_wide_dt['exog_1'].copy()
}

THIS_DIR = Path(__file__).parent
series_dict_nans = joblib.load(THIS_DIR/'fixture_sample_multi_series.joblib')
exog_dict_nans = joblib.load(THIS_DIR/'fixture_sample_multi_series_exog.joblib')

for k in series_dict_nans.keys():
    series_dict_nans[k].index.name = None
for k in exog_dict_nans.keys():
    exog_dict_nans[k].index.name = None

end_train = "2016-07-31 23:59:00"
series_dict_nans_train = {k: v.loc[:end_train,] for k, v in series_dict_nans.items()}
exog_dict_nans_train = {k: v.loc[:end_train,] for k, v in exog_dict_nans.items()}
series_dict_nans_test = {k: v.loc[end_train:,] for k, v in series_dict_nans.items()}
exog_dict_nans_test = {k: v.loc[end_train:,] for k, v in exog_dict_nans.items()}


def expected_df_to_long_format(
    df: pd.DataFrame, method: str = "predict"
) -> pd.DataFrame:
    """
    Convert DataFrame with predictions (one column per level) to long format.
    """

    if method == "predict":
        df = (
            df.melt(var_name="level", value_name="pred", ignore_index=False)
            .reset_index()
            .sort_values(by=["index", "level"])
            .set_index("index")
            .rename_axis(None, axis=0)
        )
    elif method == "bootstrapping":
        df = (
            pd.concat([value.assign(level=key) for key, value in df.items()])
            .reset_index()
            .sort_values(by=["index", "level"])
            .set_index("index")
            .rename_axis(None, axis=0)
        )
        df = df[
            ["level"] + [col for col in df.columns if col not in ["level", "index"]]
        ]
        if isinstance(df.index, pd.DatetimeIndex) and df.index.freq is not None:
            df.index.freq = None
    elif method == "interval":
        df = df.melt(var_name="level", value_name="pred", ignore_index=False).reset_index()
        df['level_aux'] = df['level'].str.replace(r'_lower_bound|_upper_bound', '', regex=True)
        df['bound_type'] = df['level'].str.extract(r'(lower_bound|upper_bound)$', expand=False).fillna('pred')

        df = (
            df.pivot_table(index=["index", "level_aux"], columns="bound_type", values="pred")
            .reset_index()
            .sort_values(by=["index", "level_aux"])
            .set_index("index")
            .rename_axis(None, axis=0)
            .rename_axis(None, axis=1)
            .rename(columns={"level_aux": "level"})
            [['level', 'pred', 'lower_bound', 'upper_bound']]
        )
    elif method == "backtesting-predict":
        df = (
            df.melt(id_vars=['fold'], var_name="level", value_name="pred", ignore_index=False)
            .reset_index()
            .sort_values(by=["index", "level"])
            .set_index("index")
            .rename_axis(None, axis=0)
            [['level', 'fold', 'pred']]
        )
    elif method == "backtesting-interval":
        df = df.melt(id_vars=['fold'], var_name="level", value_name="pred", ignore_index=False).reset_index()
        df['level_aux'] = df['level'].str.replace(r'_lower_bound|_upper_bound', '', regex=True)
        df['bound_type'] = df['level'].str.extract(r'(lower_bound|upper_bound)$', expand=False).fillna('pred')

        df = (
            df.pivot_table(index=["index", "level_aux", "fold"], columns="bound_type", values="pred")
            .reset_index()
            .sort_values(by=["index", "level_aux"])
            .set_index("index")
            .rename_axis(None, axis=0)
            .rename_axis(None, axis=1)
            .rename(columns={"level_aux": "level"})
            [['level', 'fold', 'pred', 'lower_bound', 'upper_bound']]
        )

    return df



################################################################################
#                            ForecasterSarimax                                 #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from __future__ import annotations
from typing_extensions import deprecated
from ..exceptions import runtime_deprecated
from ._forecaster_stats import ForecasterStats


# TODO: Remove in version 0.20.0
@runtime_deprecated(replacement="ForecasterStats", version="0.19.0", removal="0.20.0")
@deprecated("`ForecasterSarimax` is deprecated since version 0.19.0; use `ForecasterStats` instead. It will be removed in version 0.20.0.")
class ForecasterSarimax():
    """
    !!! warning "Deprecated"
        This class is deprecated since skforecast 0.19. Please use
        `skforecast.recursive.ForecasterStats` instead.
    
    """
    
    def __new__(
        self,
        estimator: object = None,
        transformer_y: object | None = None,
        transformer_exog: object | None = None,
        fit_kwargs: dict[str, object] | None = None,
        forecaster_id: str | int | None = None,
        regressor: object = None
    ) -> None:
        
        return ForecasterStats(
            estimator=estimator,
            transformer_y=transformer_y,
            transformer_exog=transformer_exog,
            fit_kwargs=fit_kwargs,
            forecaster_id=forecaster_id,
            regressor=regressor
        )



# Unit test set_fit_kwargs ForecasterRecursive
# ==============================================================================
from skforecast.recursive import ForecasterRecursive
from lightgbm import LGBMRegressor


def test_set_fit_kwargs():
    """
    Test set_fit_kwargs method.
    """
    forecaster = ForecasterRecursive(
                     estimator  = LGBMRegressor(),
                     lags       = 3,
                     fit_kwargs = {'categorical_feature': 'auto'}
                 )
    
    new_fit_kwargs = {'categorical_feature': ['exog']}
    forecaster.set_fit_kwargs(new_fit_kwargs)
    results = forecaster.fit_kwargs

    expected = {'categorical_feature': ['exog']}

    assert results == expected


# Unit test set_in_sample_residuals ForecasterRecursiveMultiSeries
# ==============================================================================
import re
import pytest
from pytest import approx
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import LinearRegression
from skforecast.recursive import ForecasterRecursiveMultiSeries

# Fixtures
from .fixtures_forecaster_recursive_multiseries import (
    series_dict_range, 
    exog_wide_range
)


def test_set_in_sample_residuals_NotFittedError_when_forecaster_not_fitted():
    """
    Test NotFittedError is raised when forecaster is not fitted.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)

    err_msg = re.escape(
        "This forecaster is not fitted yet. Call `fit` with appropriate "
        "arguments before using `set_in_sample_residuals()`."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.set_in_sample_residuals(series=series_dict_range)


@pytest.mark.parametrize("diff_index", 
                         [pd.RangeIndex(start=50, stop=100), 
                          pd.date_range(start='1991-07-01', periods=50, freq='MS')], 
                         ids=lambda idx: f'diff_index: {idx[[0, -1]]}')
def test_set_in_sample_residuals_IndexError_when_series_has_different_index_than_training(diff_index):
    """
    Test IndexError is raised when series has different index than training.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series_dict_range, exog=exog_wide_range['exog_1'])

    series_diff_index = {
        'l1': pd.Series(np.arange(50), name='l1'),
        'l2': pd.Series(np.arange(50), name='l2'),
    }
    series_diff_index['l1'].index = diff_index
    series_diff_index['l2'].index = diff_index
    series_diff_index_range = series_diff_index['l1'].index[[0, -1]]

    exog_diff_index = exog_wide_range.copy()
    exog_diff_index.index = diff_index

    err_msg = re.escape(
        f"The index range for series 'l1' does not match the range "
        f"used during training. Please ensure the index is aligned "
        f"with the training data.\n"
        f"    Expected : {forecaster.training_range_['l1']}\n"
        f"    Received : {series_diff_index_range}"
    )
    with pytest.raises(IndexError, match = err_msg):
        forecaster.set_in_sample_residuals(
            series=series_diff_index, exog=exog_diff_index['exog_1']
        )


def test_set_in_sample_residuals_ValueError_when_X_train_features_names_out_not_the_same():
    """
    Test ValueError is raised when X_train_features_names_out are different from 
    the ones used in training.
    """
    forecaster = ForecasterRecursiveMultiSeries(LinearRegression(), lags=3)
    forecaster.fit(series=series_dict_range, exog=exog_wide_range['exog_1'])

    err_msg = re.escape(
        "Feature mismatch detected after matrix creation. The features "
        "generated from the provided data do not match those used during "
        "the training process. To correctly set in-sample residuals, "
        "ensure that the same data and preprocessing steps are applied.\n"
        "    Expected output : ['lag_1', 'lag_2', 'lag_3', '_level_skforecast', 'exog_1']\n"
        "    Current output  : ['lag_1', 'lag_2', 'lag_3', '_level_skforecast']"
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_in_sample_residuals(series=series_dict_range)


@pytest.mark.parametrize("encoding", 
                         ["ordinal", "ordinal_category", "onehot", None],
                         ids = lambda value: f'encoding: {value}')
def test_set_in_sample_residuals_store_same_residuals_as_fit(encoding):
    """
    Test that set_in_sample_residuals stores same residuals as fit.
    """
    forecaster_1 = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, transformer_series=StandardScaler(), 
        differentiation=1, binner_kwargs={'n_bins': 3}
    )
    forecaster_1.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=True
    )
    results_residuals_1 = forecaster_1.in_sample_residuals_
    results_residuals_bin_1 = forecaster_1.in_sample_residuals_by_bin_
    results_binner_intervals_1 = forecaster_1.binner_intervals_

    forecaster_2 = ForecasterRecursiveMultiSeries(
        LinearRegression(), lags=3, encoding=encoding, transformer_series=StandardScaler(), 
        differentiation=1, binner_kwargs={'n_bins': 3}
    )
    forecaster_2.fit(
        series=series_dict_range, exog=exog_wide_range['exog_1'], store_in_sample_residuals=False
    )
    scaler_id_after_fit = {
        level: id(scaler) for level, scaler in forecaster_2.transformer_series_.items()
    }
    differentiator_id_after_fit = {
        level: id(differentiator) for level, differentiator in forecaster_2.differentiator_.items()
    }
    forecaster_2.set_in_sample_residuals(
        series=series_dict_range, exog=exog_wide_range['exog_1']
    )
    scaler_id_after_set_in_sample_residuals = {
        level: id(scaler) for level, scaler in forecaster_2.transformer_series_.items()
    }
    differentiator_id_after_set_in_sample_residuals = {
        level: id(differentiator) for level, differentiator in forecaster_2.differentiator_.items()
    }
    results_residuals_2 = forecaster_2.in_sample_residuals_
    results_residuals_bin_2 = forecaster_2.in_sample_residuals_by_bin_
    results_binner_intervals_2 = forecaster_2.binner_intervals_

    # Attributes
    assert forecaster_1.series_names_in_ == forecaster_2.series_names_in_
    assert forecaster_1.X_train_series_names_in_ == forecaster_2.X_train_series_names_in_
    assert forecaster_1.X_train_window_features_names_out_ == forecaster_2.X_train_window_features_names_out_
    assert forecaster_1.X_train_features_names_out_ == forecaster_2.X_train_features_names_out_

    # Transformers
    for level in scaler_id_after_fit.keys():
        assert scaler_id_after_fit[level] == scaler_id_after_set_in_sample_residuals[level]

    # Differentiators
    for level in differentiator_id_after_fit.keys():
        assert differentiator_id_after_fit[level] == differentiator_id_after_set_in_sample_residuals[level]

    # In-sample residuals
    assert results_residuals_1.keys() == results_residuals_2.keys()
    for k in results_residuals_1.keys():
        np.testing.assert_array_almost_equal(
            np.sort(results_residuals_1[k]), np.sort(results_residuals_2[k])
        )

    # In-sample residuals by bin
    assert results_residuals_bin_1.keys() == results_residuals_bin_2.keys()
    for level in results_residuals_bin_1.keys():
        assert results_residuals_bin_1[level].keys() == results_residuals_bin_2[level].keys()
        for k in results_residuals_bin_1[level].keys():
            np.testing.assert_array_almost_equal(
                results_residuals_bin_1[level][k], results_residuals_bin_2[level][k]
            )
    
    # Binner intervals
    assert results_binner_intervals_1.keys() == results_binner_intervals_2.keys()
    for level in results_binner_intervals_1.keys():
        assert results_binner_intervals_1[level].keys() == results_binner_intervals_2[level].keys()
        for k in results_binner_intervals_1[level].keys():
            assert results_binner_intervals_1[level][k][0] == approx(results_binner_intervals_2[level][k][0])
            assert results_binner_intervals_1[level][k][1] == approx(results_binner_intervals_2[level][k][1])



# Unit test __init__ ForecasterRecursive
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.recursive import ForecasterRecursive


def test_init_ValueError_when_no_lags_or_window_features():
    """
    Test ValueError is raised when no lags or window_features are passed.
    """
    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursive(
            estimator       = LinearRegression(),
            lags            = None,
            window_features = None
        )


@pytest.mark.parametrize("lags, window_features, expected", 
                         [(5, None, 5), 
                          (None, True, 6), 
                          ([], True, 6), 
                          (5, True, 6)], 
                         ids = lambda dt: f'lags, window_features, expected: {dt}')
def test_init_window_size_correctly_stored(lags, window_features, expected):
    """
    Test window_size is correctly stored when lags or window_features are passed.
    """
    if window_features:
        window_features = RollingFeatures(
            stats=['ratio_min_max', 'median'], window_sizes=[5, 6]
        )

    forecaster = ForecasterRecursive(
                     estimator       = LinearRegression(),
                     lags            = lags,
                     window_features = window_features
                 )
    
    assert forecaster.window_size == expected
    if lags:
        np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
        assert forecaster.lags_names == [f'lag_{i}' for i in range(1, lags + 1)]
        assert forecaster.max_lag == lags
    else:
        assert forecaster.lags is None
        assert forecaster.lags_names is None
        assert forecaster.max_lag is None
    if window_features:
        assert forecaster.window_features_names == ['roll_ratio_min_max_5', 'roll_median_6']
        assert forecaster.window_features_class_names == ['RollingFeatures']
    else:
        assert forecaster.window_features_names is None
        assert forecaster.window_features_class_names is None


@pytest.mark.parametrize("dif", 
                         [0, 0.5, 1.5, 'not_int'], 
                         ids = lambda dif: f'differentiation: {dif}')
def test_init_ValueError_when_differentiation_argument_is_not_int_or_greater_than_0(dif):
    """
    Test ValueError is raised when differentiation is not an int or greater than 0.
    """
    err_msg = re.escape(
        f"Argument `differentiation` must be an integer equal to or "
        f"greater than 1. Got {dif}."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursive(
            estimator       = LinearRegression(),
            lags            = 5,
            differentiation = dif
        )


@pytest.mark.parametrize("dif", 
                         [1, 2], 
                         ids = lambda dif: f'differentiation: {dif}')
def test_init_window_size_is_increased_when_differentiation(dif):
    """
    Test window_size is increased when including differentiation.
    """
    forecaster = ForecasterRecursive(
                     estimator       = LinearRegression(),
                     lags            = 5,
                     differentiation = dif
                 )
    
    assert forecaster.differentiation == dif
    assert forecaster.differentiation_max == dif
    assert forecaster.window_size == len(forecaster.lags) + dif


def test_init_binner_is_created_when_binner_kwargs_is_None():
    """
    Test binner is initialized with the default kwargs
    """
    forecaster = ForecasterRecursive(
                     estimator = object(),
                     lags      = 5,
                 )
    
    expected = {
        'n_bins': 10, 'method': 'linear', 'subsample': 200000,
        'random_state': 789654, 'dtype': np.float64
    }

    assert forecaster.binner.get_params() == expected


def test_init_binner_is_created_when_binner_kwargs_is_not_None():
    """
    Test binner is initialized with kwargs
    """
    binner_kwargs = {
        'n_bins': 10, 'method': 'linear', 'subsample': 500,
        'random_state': 1234, 'dtype': np.float64
    }
    forecaster = ForecasterRecursive(
                     estimator     = object(),
                     lags          = 5,
                     binner_kwargs = binner_kwargs
                 )
    
    expected = {
        'n_bins': 10, 'method': 'linear', 'subsample': 500,
        'random_state': 1234, 'dtype': np.float64
    }

    assert forecaster.binner.get_params() == expected



# Unit test fit ForecasterDirectMultiVariate
# ==============================================================================
import pytest
from pytest import approx
import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirectMultiVariate

# Fixtures
from .fixtures_forecaster_direct_multivariate import series as series_fixtures
from .fixtures_forecaster_direct_multivariate import exog

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )


def test_forecaster_series_exog_features_stored():
    """
    Test forecaster stores series and exog features after fitting.
    """
    rolling = RollingFeatures(
        stats=['ratio_min_max', 'median'], window_sizes=4
    )
    forecaster = ForecasterDirectMultiVariate(
                     estimator        = LinearRegression(), 
                     level            = 'l1',
                     steps            = 2,
                     lags             = 3,
                     window_features  = rolling,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(series=series_fixtures, exog=exog)

    series_names_in_ = ['l1', 'l2']
    exog_in_ = True
    exog_type_in_ = type(exog)
    exog_names_in_ = ['exog_1', 'exog_2']
    exog_dtypes_in_ = {'exog_1': exog['exog_1'].dtype, 'exog_2': exog['exog_2'].dtype}
    # All floats
    exog_dtypes_out_ = {'exog_1': exog['exog_1'].dtype, 'exog_2_a': exog['exog_1'].dtype, 'exog_2_b': exog['exog_1'].dtype}
    X_train_series_names_in_ = ['l1', 'l2']
    X_train_window_features_names_out_ = [
        'l1_roll_ratio_min_max_4', 'l1_roll_median_4',
        'l2_roll_ratio_min_max_4', 'l2_roll_median_4'
    ]
    X_train_exog_names_out_ = ['exog_1', 'exog_2_a', 'exog_2_b']
    X_train_direct_exog_names_out_ = [
        'exog_1_step_1', 'exog_2_a_step_1', 'exog_2_b_step_1',
        'exog_1_step_2', 'exog_2_a_step_2', 'exog_2_b_step_2'
    ]
    X_train_features_names_out_ = [
        'l1_lag_1', 'l1_lag_2', 'l1_lag_3', 'l1_roll_ratio_min_max_4', 'l1_roll_median_4',
        'l2_lag_1', 'l2_lag_2', 'l2_lag_3', 'l2_roll_ratio_min_max_4', 'l2_roll_median_4',
        'exog_1_step_1', 'exog_2_a_step_1', 'exog_2_b_step_1', 
        'exog_1_step_2', 'exog_2_a_step_2', 'exog_2_b_step_2'
    ]
    
    assert forecaster.series_names_in_ == series_names_in_
    assert forecaster.exog_in_ == exog_in_
    assert forecaster.exog_type_in_ == exog_type_in_
    assert forecaster.exog_names_in_ == exog_names_in_
    assert forecaster.exog_dtypes_in_ == exog_dtypes_in_
    assert forecaster.exog_dtypes_out_ == exog_dtypes_out_
    assert forecaster.X_train_series_names_in_ == X_train_series_names_in_
    assert forecaster.X_train_window_features_names_out_ == X_train_window_features_names_out_
    assert forecaster.X_train_exog_names_out_ == X_train_exog_names_out_
    assert forecaster.X_train_direct_exog_names_out_ == X_train_direct_exog_names_out_
    assert forecaster.X_train_features_names_out_ == X_train_features_names_out_


def test_fit_correct_dict_create_transformer_series():
    """
    Test fit method creates correctly all the auxiliary dicts transformer_series_.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10)), 
                           'l3': pd.Series(np.arange(10))})

    transformer_series = {'l1': StandardScaler(), 'l3': StandardScaler(), 'l4': StandardScaler()}

    forecaster = ForecasterDirectMultiVariate(
                     estimator          = LinearRegression(), 
                     level              = 'l1',
                     lags               = 3,
                     steps              = 2,
                     transformer_series = transformer_series
                 )
    
    forecaster.fit(series=series, store_in_sample_residuals=False)
    expected_transformer_series_ = {'l1': forecaster.transformer_series_['l1'], 'l2': None, 'l3': forecaster.transformer_series_['l3']}

    assert forecaster.transformer_series_ == expected_transformer_series_

    forecaster.fit(series=series[['l1', 'l2']], store_in_sample_residuals=False)
    expected_transformer_series_ = {'l1': forecaster.transformer_series_['l1'], 'l2': None}

    assert forecaster.transformer_series_ == expected_transformer_series_


def test_forecaster_DatetimeIndex_index_freq_stored():
    """
    Test series.index.freq is stored in forecaster.index_freq.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})

    series.index = pd.date_range(start='2022-01-01', periods=10, freq='1D')

    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=1
    )
    forecaster.fit(series=series)
    expected = series.index.freq
    results = forecaster.index_freq_

    assert results == expected


def test_forecaster_index_step_stored():
    """
    Test serie without DatetimeIndex, step is stored in forecaster.index_freq_.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})
    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2
    )
    forecaster.fit(series=series)
    expected = series.index.step
    results = forecaster.index_freq_

    assert results == expected
    

@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_fit_in_sample_residuals_stored(n_jobs):
    """
    Test that values of in_sample_residuals_ are stored after fitting.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(10))})
    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2, n_jobs=n_jobs
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    expected = {
        "l1": np.array(
                [0.0000000e+00, 0.0000000e+00, 0.0000000e+00, 
                 0.0000000e+00, 0.0000000e+00, 8.8817842e-16, 0., 0., 0., 0., 0., 0.]
              )
    }
    results = forecaster.in_sample_residuals_

    assert isinstance(results, dict)
    assert results.keys() == expected.keys()
    for k in expected.keys():
        assert isinstance(results[k], np.ndarray)
        np.testing.assert_array_almost_equal(results[k], expected[k])


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_fit_same_residuals_when_residuals_greater_than_10000(n_jobs):
    """
    Test fit return same residuals when residuals len is greater than 10_000.
    Testing with two different forecaster.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(12_000)), 
                           'l2': pd.Series(np.arange(12_000))})
    
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2, n_jobs=n_jobs
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_1 = forecaster.in_sample_residuals_

    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2, n_jobs=n_jobs
    )
    forecaster.fit(series=series, store_in_sample_residuals=True)
    results_2 = forecaster.in_sample_residuals_

    assert isinstance(results_1, dict)
    assert isinstance(results_2, dict)
    assert results_1.keys() == results_2.keys()
    for k in results_1.keys():
        assert isinstance(results_1[k], np.ndarray)
        assert isinstance(results_2[k], np.ndarray)
        assert len(results_1[k]) == 10_000
        assert len(results_2[k]) == 10_000
        np.testing.assert_array_equal(results_1[k], results_2[k])


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_fit_in_sample_residuals_by_bin_stored(n_jobs):
    """
    Test that values of in_sample_residuals_by_bin are stored after fitting.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2, binner_kwargs={'n_bins': 3}, n_jobs=n_jobs
    )
    forecaster.fit(series=series_fixtures, store_in_sample_residuals=True)

    expected_1 = {
        'l1' : np.array([
                    0.61351689,  0.71846659, -0.46065047,  1.8740517 ,  0.35951685,
                    0.36635782, -0.3225934 , -0.86465358,  0.82252715, -0.2741012 ,
                    -1.70423914, -0.33664349,  1.16033533, -1.66552242, -1.26639479,
                    0.18116919, -0.35795787,  1.02694562,  1.33431561,  0.98054222,
                    0.94382542,  0.57275958, -1.52361509, -0.76119268, -0.84770863,
                    -1.07296309,  0.20287787, -1.30813619,  0.3000927 , -0.59338942,
                    0.52690419, -0.5786443 , -0.43975004,  0.23252597,  1.25096008,
                    1.05277835, -0.17913321,  0.65860599, -1.73952616,  0.00926585,
                    -0.74486859,  1.56624963, -0.89792655,  0.1841916 ,  1.34397847,
                    -0.34315036, 0.89701176, -0.41773024,  1.82467126,  0.72112853,  0.40786125,
                    -0.61189293, -0.60896505,  0.65160718, -0.30338729, -1.93442134,
                    -0.79019702,  0.95739893, -1.24036132, -1.99937927,  0.22147022,
                    0.12771388,  0.72032178,  1.67692078,  0.86091935,  0.9791868 ,
                    0.59836304, -0.8500661 , -0.88860757, -0.87948968, -0.67394226,
                    0.29203755, -1.50705801,  0.1996901 , -0.79372163,  0.47123556,
                    -0.65428718, -0.94038195, -0.16313875,  1.38627477,  1.49653973,
                    -0.01688572,  0.7562465 , -1.50022778, -0.32925933, -0.60151835,
                    1.26646515, -1.08437562, -0.21536791,  1.62120186,  0.11599908,
                    0.75439722
                ])
    }

    expected_2 = {
        'l1': {
            0: np.array([0.61351689,  0.36635782, -0.3225934 , -1.70423914,  1.16033533,
                    -1.26639479,  1.02694562,  0.94382542, -0.84770863, -1.30813619,
                    0.3000927 ,  0.52690419, -0.43975004,  0.23252597,  0.00926585,
                    -0.89792655,  0.1841916 ,  0.40786125,  0.72032178,  1.67692078,
                    0.9791868 , -0.87948968, -0.67394226, -1.50705801,  0.1996901 ,
                    0.47123556, -0.16313875,  0.7562465 , -1.50022778, -0.32925933,
                    0.75439722]),
            1: np.array([-0.46065047,  0.82252715, -0.2741012 , -0.33664349,  0.18116919,
                    1.33431561,  0.98054222, -0.76119268, -1.07296309,  0.65860599,
                    -1.73952616,  1.56624963,  0.89701176, -0.41773024,  0.72112853,
                    -0.61189293, -0.60896505, -0.30338729, -1.93442134,  0.95739893,
                    -1.24036132,  0.22147022,  0.12771388,  0.86091935, -0.8500661 ,
                    -0.94038195, -0.01688572, -1.08437562, -0.21536791,  0.11599908]),
            2: np.array([ 0.71846659,  1.8740517 ,  0.35951685, -0.86465358, -1.66552242,
                    -0.35795787,  0.57275958, -1.52361509,  0.20287787, -0.59338942,
                    -0.5786443 ,  1.25096008,  1.05277835, -0.17913321, -0.74486859,
                    1.34397847, -0.34315036,  1.82467126,  0.65160718, -0.79019702,
                    -1.99937927,  0.59836304, -0.88860757,  0.29203755, -0.79372163,
                    -0.65428718,  1.38627477,  1.49653973, -0.60151835,  1.26646515,
                    1.62120186])
        }
    }

    expected_3 = {
        'l1': {
            0: (-0.8032116858231456, -0.13623943835630237),
            1: (-0.13623943835630237, 0.17856930169525217),
            2: (0.17856930169525217, 0.8498028419581619)
        }
    }

    np.testing.assert_array_almost_equal(forecaster.in_sample_residuals_['l1'], expected_1['l1'])
    for k in forecaster.in_sample_residuals_by_bin_['l1'].keys():
        np.testing.assert_array_almost_equal(
            forecaster.in_sample_residuals_by_bin_['l1'][k], expected_2['l1'][k]
        )
    for k in forecaster.binner_intervals_['l1'].keys():
        assert forecaster.binner_intervals_['l1'][k][0] == approx(expected_3['l1'][k][0])
        assert forecaster.binner_intervals_['l1'][k][1] == approx(expected_3['l1'][k][1])


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_fit_in_sample_residuals_not_stored_probabilistic_mode_binned(n_jobs):
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False`. Binner intervals are stored.
    """
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2, binner_kwargs={'n_bins': 3}, n_jobs=n_jobs
    )
    forecaster.fit(series=series_fixtures, store_in_sample_residuals=False)

    expected = {'l1': None}
    expected_binner_intervals_ = {
        'l1' :{
            0: (-0.8032116858231456, -0.13623943835630237),
            1: (-0.13623943835630237, 0.17856930169525217),
            2: (0.17856930169525217, 0.8498028419581619)
        }
    }

    assert isinstance(forecaster.in_sample_residuals_, dict)
    assert forecaster.in_sample_residuals_.keys() == expected.keys()
    for k in forecaster.in_sample_residuals_.keys():
        assert forecaster.in_sample_residuals_[k] == expected[k]
    
    assert forecaster.in_sample_residuals_by_bin_ == {'l1': None}

    assert forecaster.binner_intervals_.keys() == expected_binner_intervals_.keys()
    for k in forecaster.binner_intervals_['l1'].keys():
        assert forecaster.binner_intervals_['l1'][k][0] == approx(expected_binner_intervals_['l1'][k][0])
        assert forecaster.binner_intervals_['l1'][k][1] == approx(expected_binner_intervals_['l1'][k][1])


@pytest.mark.parametrize("n_jobs", [1, -1, 'auto'], 
                         ids=lambda n_jobs: f'n_jobs: {n_jobs}')
def test_fit_in_sample_residuals_not_stored_probabilistic_mode_False(n_jobs):
    """
    Test that values of in_sample_residuals_ are not stored after fitting
    when `store_in_sample_residuals=False` and _probabilistic_mode=False.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(5)), 
                           'l2': pd.Series(np.arange(5))})
    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2, binner_kwargs={'n_bins': 3}, n_jobs=n_jobs
    )
    forecaster._probabilistic_mode = False
    forecaster.fit(series=series, store_in_sample_residuals=False)
    
    expected = {'l1': None}

    assert isinstance(forecaster.in_sample_residuals_, dict)
    assert forecaster.in_sample_residuals_.keys() == expected.keys()
    for k in forecaster.in_sample_residuals_.keys():
        assert forecaster.in_sample_residuals_[k] == expected[k]
    assert forecaster.in_sample_residuals_by_bin_ == {'l1': None}
    assert forecaster.binner_intervals_ == {}


@pytest.mark.parametrize("store_last_window", 
                         [True, ['l1', 'l2'], False], 
                         ids=lambda lw: f'store_last_window: {lw}')
def test_fit_last_window_stored(store_last_window):
    """
    Test that values of last window are stored after fitting.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(50, 60))})

    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l1', lags=3, steps=2
    )
    forecaster.fit(series=series, store_last_window=store_last_window)

    expected = pd.DataFrame({
        'l1': pd.Series(np.array([7, 8, 9])), 
        'l2': pd.Series(np.array([57, 58, 59]))
    })
    expected.index = pd.RangeIndex(start=7, stop=10, step=1)

    if store_last_window:
        pd.testing.assert_frame_equal(forecaster.last_window_, expected)
        assert forecaster.series_names_in_ == ['l1', 'l2']
        assert forecaster.X_train_series_names_in_ == ['l1', 'l2']
    else:
        assert forecaster.last_window_ is None


def test_fit_last_window_stored_when_different_lags():
    """
    Test that values of last window are stored after fitting when different lags
    configurations.
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(100, 110))})

    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level='l2', steps = 2, lags = {'l1': 3, 'l2': [1, 5]}
    )
    forecaster.fit(series=series)

    expected = pd.DataFrame({
        'l1': pd.Series(np.array([5, 6, 7, 8, 9])), 
        'l2': pd.Series(np.array([105, 106, 107, 108, 109]))
    })
    expected.index = pd.RangeIndex(start=5, stop=10, step=1)

    pd.testing.assert_frame_equal(forecaster.last_window_, expected)
    assert forecaster.series_names_in_ == ['l1', 'l2']
    assert forecaster.X_train_series_names_in_ == ['l1', 'l2']


@pytest.mark.parametrize("level",
                         ['l1', 'l2'],
                         ids=lambda level: f'level: {level}')
def test_fit_last_window_stored_when_lags_dict_with_None(level):
    """
    Test that values of last window are stored after fitting when lags is a dict
    with None values.    
    """
    series = pd.DataFrame({'l1': pd.Series(np.arange(10)), 
                           'l2': pd.Series(np.arange(100, 110))})

    forecaster = ForecasterDirectMultiVariate(
        estimator=LinearRegression(), level=level, steps = 2, lags = {'l1': 3, 'l2': None}
    )
    forecaster.fit(series=series)

    expected = pd.DataFrame({'l1': pd.Series(np.array([7, 8, 9]))})
    expected.index = pd.RangeIndex(start=7, stop=10, step=1)

    pd.testing.assert_frame_equal(forecaster.last_window_, expected)
    assert forecaster.series_names_in_ == ['l1', 'l2']
    assert forecaster.X_train_series_names_in_ == ['l1']



################################################################################
#                                   Plot                                       #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from __future__ import annotations
import os
from typing import Any
from pathlib import Path
import warnings
import numpy as np
import pandas as pd
from ..utils import check_optional_dependency, input_to_frame
from ..exceptions import IgnoredArgumentWarning

try:
    import matplotlib
    import matplotlib.pyplot as plt
    from matplotlib.animation import FuncAnimation, PillowWriter
    import seaborn as sns
    from statsmodels.graphics.tsaplots import plot_acf
    from statsmodels.tsa.stattools import acf, pacf
except Exception as e:
    package_name = str(e).split(" ")[-1].replace("'", "")
    check_optional_dependency(package_name=package_name)


def plot_residuals(
    residuals: np.ndarray | pd.Series | None = None,
    y_true: np.ndarray | pd.Series | None = None,
    y_pred: np.ndarray | pd.Series | None = None,
    fig: matplotlib.figure.Figure | None = None,
    **fig_kw
) -> matplotlib.figure.Figure:
    """
    Parameters
    ----------
    residuals : pandas Series, numpy ndarray, default None.
        Values of residuals. If `None`, residuals are calculated internally using
        `y_true` and `y_true`.
    y_true : pandas Series, numpy ndarray, default None.
        Ground truth (correct) values. Ignored if residuals is not `None`.
    y_pred : pandas Series, numpy ndarray, default None. 
        Values of predictions. Ignored if residuals is not `None`.
    fig : matplotlib.figure.Figure, default None. 
        Pre-existing fig for the plot. Otherwise, call matplotlib.pyplot.figure()
        internally.
    fig_kw : dict
        Other keyword arguments are passed to matplotlib.pyplot.figure()

    Returns
    -------
    fig: matplotlib.figure.Figure
        Matplotlib Figure.
    
    """
    
    if residuals is None and (y_true is None or y_pred is None):
        raise ValueError(
            "If `residuals` argument is None then, `y_true` and `y_pred` must be provided."
        )
        
    if residuals is None:
        residuals = y_true - y_pred
            
    if fig is None:
        fig = plt.figure(constrained_layout=True, **fig_kw)
        
    gs  = matplotlib.gridspec.GridSpec(2, 2, figure=fig)
    ax1 = plt.subplot(gs[0, :])
    ax2 = plt.subplot(gs[1, 0])
    ax3 = plt.subplot(gs[1, 1])
    
    ax1.plot(residuals)
    sns.histplot(residuals, kde=True, bins=30, ax=ax2)
    plot_acf(residuals, ax=ax3, lags=60)
    
    ax1.set_title("Residuals")
    ax2.set_title("Distribution")
    ax3.set_title("Autocorrelation")

    return fig


def plot_multivariate_time_series_corr(
    corr: pd.DataFrame,
    ax: matplotlib.axes.Axes | None = None,
    **fig_kw
) -> matplotlib.figure.Figure:
    """
    Heatmap plot of a correlation matrix.

    Parameters
    ----------
    corr : pandas DataFrame
        correlation matrix
    ax : matplotlib.axes.Axes, default None
        Pre-existing ax for the plot. Otherwise, call matplotlib.pyplot.subplots() 
        internally.
    fig_kw : dict
        Other keyword arguments are passed to matplotlib.pyplot.subplots()
    
    Returns
    -------
    fig: matplotlib.figure.Figure
        Matplotlib Figure.

    """

    if ax is None:
        fig, ax = plt.subplots(1, 1, **fig_kw)
    
    sns.heatmap(
        corr,
        annot=True,
        linewidths=.5,
        ax=ax,
        cmap=sns.color_palette("viridis", as_cmap=True)
    )

    ax.set_xlabel('Time series')
    
    return fig


def plot_prediction_distribution(
    bootstrapping_predictions: pd.DataFrame,
    bw_method: Any | None = None,
    **fig_kw
) -> matplotlib.figure.Figure:
    """
    Ridge plot of bootstrapping predictions. This plot is very useful to understand 
    the uncertainty of forecasting predictions.

    Parameters
    ----------
    bootstrapping_predictions : pandas DataFrame
        Bootstrapping predictions created with `Forecaster.predict_bootstrapping`.
    bw_method : str, scalar, Callable, default None
        The method used to calculate the estimator bandwidth. This can be 'scott', 
        'silverman', a scalar constant or a Callable. If None (default), 'scott' 
        is used. See scipy.stats.gaussian_kde for more information.
    fig_kw : dict
        All additional keyword arguments are passed to the `pyplot.figure` call.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Matplotlib Figure.
    
    """

    index = bootstrapping_predictions.index.astype(str).to_list()[::-1]
    palette = sns.cubehelix_palette(len(index), rot=-.25, light=.7, reverse=False)
    fig, axs = plt.subplots(len(index), 1, sharex=True, **fig_kw)
    if not isinstance(axs, np.ndarray):
        axs = np.array([axs])

    for i, step in enumerate(index):
        plot = (
            bootstrapping_predictions.loc[step, :]
            .plot.kde(ax=axs[i], bw_method=bw_method, lw=0.5)
        )

        # Fill density area
        x = plot.get_children()[0]._x
        y = plot.get_children()[0]._y
        axs[i].fill_between(x, y, color=palette[i])
        prediction_mean = bootstrapping_predictions.loc[step, :].mean()
        
        # Closest point on x to the prediction mean
        idx = np.abs(x - prediction_mean).argmin()
        axs[i].vlines(x[idx], ymin=0, ymax=y[idx], linestyle="dashed", color='w')

        axs[i].spines['top'].set_visible(False)
        axs[i].spines['right'].set_visible(False)
        axs[i].spines['bottom'].set_visible(False)
        axs[i].spines['left'].set_visible(False)
        axs[i].set_yticklabels([])
        axs[i].set_yticks([])
        axs[i].set_ylabel(step, rotation='horizontal')
        axs[i].set_xlabel('prediction')

    fig.subplots_adjust(hspace=-0)
    fig.suptitle('Forecasting distribution per step')

    return fig


def set_dark_theme(
    custom_style: dict | None = None
) -> None:
    """
    Set aspects of the visual theme for all matplotlib plots.
    This function changes the global defaults for all plots using the matplotlib
    rcParams system. The theme includes specific colors for figure and axes
    backgrounds, gridlines, text, labels, and ticks. It also sets the font size
    and line width.

    Parameters
    ----------
    custom_style : dict, default None
        Optional dictionary containing custom styles to be added or override the
        default dark theme. It is applied after the default theme is set by
        using the `plt.rcParams.update()` method.

    Returns
    -------
    None
    
    """

    plt.style.use('fivethirtyeight')
    dark_style = {
        'figure.facecolor': '#001633',
        'axes.facecolor': '#001633',
        'savefig.facecolor': '#001633',
        'axes.grid': True,
        'axes.grid.which': 'both',
        'axes.spines.left': False,
        'axes.spines.right': False,
        'axes.spines.top': False,
        'axes.spines.bottom': False,
        'grid.color': '#212946',
        'grid.linewidth': '1',
        'text.color': '0.9',
        'axes.labelcolor': '0.9',
        'xtick.color': '0.9',
        'ytick.color': '0.9',
        'font.size': 10,
        'lines.linewidth': 1.5
    }

    if custom_style is not None:
        dark_style.update(custom_style)
        
    plt.rcParams.update(dark_style)


def plot_prediction_intervals(
    predictions: pd.DataFrame,
    y_true: pd.Series | pd.DataFrame,
    target_variable: str,
    initial_x_zoom: list[str] | None = None,
    title: str | None = None,
    xaxis_title: str | None = None,
    yaxis_title: str | None = None,
    ax: plt.Axes | None = None,
    kwargs_subplots: dict[str, object] = {'figsize': (7, 3)},
    kwargs_fill_between: dict[str, object] = {'color': '#444444', 'alpha': 0.3, 'zorder': 1}
):
    """
    Plot predicted intervals vs real values using matplotlib.

    Parameters
    ----------
    predictions : pandas DataFrame
        Predicted values and intervals. Expected columns are 'pred', 'lower_bound'
        and 'upper_bound'.
    y_true : pandas Series, pandas DataFrame
        Real values of target variable.
    target_variable : str
        Name of target variable.
    initial_x_zoom : list, default None
        Initial zoom of x-axis, by default None.
    title : str, default None
        Title of the plot, by default None.
    xaxis_title : str, default None
        Title of x-axis, by default None.
    yaxis_title : str, default None
        Title of y-axis, by default None.
    ax : matplotlib axes, default None
        Axes where to plot, by default None.
    kwargs_subplots : dict, default {'figsize': (7, 3)}
        Additional keyword arguments (key, value mappings) to pass to `plt.subplots`.
    kwargs_fill_between : dict, default {'color': '#444444', 'alpha': 0.3}
        Additional keyword arguments (key, value mappings) to pass to `ax.fill_between`.

    Returns
    -------
    None
    
    """
    
    if ax is None:
        fig, ax = plt.subplots(**kwargs_subplots)

    if isinstance(y_true, pd.Series):
        y_true = y_true.to_frame()

    y_true.loc[predictions.index, target_variable].plot(ax=ax, label='real value')
    predictions['pred'].plot(ax=ax, label='prediction')
    ax.fill_between(
        predictions.index,
        predictions['lower_bound'],
        predictions['upper_bound'],
        label='prediction interval',
        **kwargs_fill_between
    )
    ax.set_ylabel(yaxis_title)
    ax.set_xlabel(xaxis_title)
    ax.set_title(title)
    ax.legend()

    if initial_x_zoom is not None:
        ax.set_xlim(initial_x_zoom)


def calculate_lag_autocorrelation(
    data: pd.Series | pd.DataFrame,
    n_lags: int = 50,
    last_n_samples: int | None = None,
    sort_by: str = "partial_autocorrelation_abs",
    acf_kwargs: dict[str, object] = {},
    pacf_kwargs: dict[str, object] = {},
) -> pd.DataFrame:
    """
    Calculate autocorrelation and partial autocorrelation for a time series.
    This is a wrapper around statsmodels.acf [1]_ and statsmodels.pacf [2]_.

    Parameters
    ----------
    data : pandas Series, pandas DataFrame
        Time series to calculate autocorrelation. If a DataFrame is provided,
        it must have exactly one column.
    n_lags : int
        Number of lags to calculate autocorrelation.
    last_n_samples : int or None, default None
        Number of most recent samples to use. If None, use the entire series. 
        Note that partial correlations can only be computed for lags up to 
        50% of the sample size. For example, if the series has 10 samples, 
        `n_lags` must be less than or equal to 5. This parameter is useful
        to speed up calculations when the series is very long.
    sort_by : str, default 'partial_autocorrelation_abs'
        Sort results by 'lag', 'partial_autocorrelation_abs', 
        'partial_autocorrelation', 'autocorrelation_abs' or 'autocorrelation'.
    acf_kwargs : dict, default {}
        Optional arguments to pass to statsmodels.tsa.stattools.acf.
    pacf_kwargs : dict, default {}
        Optional arguments to pass to statsmodels.tsa.stattools.pacf.

    Returns
    -------
    results : pandas DataFrame
        Autocorrelation and partial autocorrelation values.

    References
    ----------
    .. [1] Statsmodels acf API Reference.
           https://www.statsmodels.org/stable/generated/statsmodels.tsa.stattools.acf.html
    
    .. [2] Statsmodels pacf API Reference.
           https://www.statsmodels.org/stable/generated/statsmodels.tsa.stattools.pacf.html

    Examples
    --------
    ```python
    import pandas as pd
    from skforecast.plot import calculate_lag_autocorrelation

    data = pd.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    calculate_lag_autocorrelation(data = data, n_lags = 4)
    
    #    lag  partial_autocorrelation_abs  partial_autocorrelation  autocorrelation_abs  autocorrelation
    # 0    1                     0.777778                 0.777778             0.700000         0.700000
    # 1    4                     0.360707                -0.360707             0.078788        -0.078788
    # 2    3                     0.274510                -0.274510             0.148485         0.148485
    # 3    2                     0.227273                -0.227273             0.412121         0.412121
    ```

    """

    if not isinstance(data, (pd.Series, pd.DataFrame)):
        raise TypeError(
            f"`data` must be a pandas Series or a DataFrame with a single column. "
            f"Got {type(data)}."
        )
    if isinstance(data, pd.DataFrame) and data.shape[1] != 1:
        raise ValueError(
            f"If `data` is a DataFrame, it must have exactly one column. "
            f"Got {data.shape[1]} columns."
        )
    if not isinstance(n_lags, int) or n_lags <= 0:
        raise TypeError(f"`n_lags` must be a positive integer. Got {n_lags}.")
    
    if last_n_samples is not None:
        if not isinstance(last_n_samples, int) or last_n_samples <= 0:
            raise TypeError(f"`last_n_samples` must be a positive integer. Got {last_n_samples}.")
        data = data.iloc[-last_n_samples:]

    if sort_by not in [
        "lag", "partial_autocorrelation_abs", "partial_autocorrelation",
        "autocorrelation_abs", "autocorrelation",
    ]:
        raise ValueError(
            "`sort_by` must be 'lag', 'partial_autocorrelation_abs', 'partial_autocorrelation', "
            "'autocorrelation_abs' or 'autocorrelation'."
        )

    pacf_values = pacf(data, nlags=n_lags, **pacf_kwargs)
    acf_values = acf(data, nlags=n_lags, **acf_kwargs)

    results = pd.DataFrame(
        {
            "lag": range(n_lags + 1),
            "partial_autocorrelation_abs": np.abs(pacf_values),
            "partial_autocorrelation": pacf_values,
            "autocorrelation_abs": np.abs(acf_values),
            "autocorrelation": acf_values,
        }
    ).iloc[1:]

    if sort_by == "lag":
        results = results.sort_values(by=sort_by, ascending=True).reset_index(drop=True)
    else:
        results = results.sort_values(by=sort_by, ascending=False).reset_index(drop=True)

    return results


def backtesting_gif_creator(
    data: pd.Series | pd.DataFrame,
    cv: object,
    series_to_plot: list[str] | None = None,
    plot_last_window: bool = False,
    filename: str = "backtesting.gif",
    plt_style: str = "ggplot",
    figsize: tuple[int, int] = (9, 4),
    colors: dict[str, str] | None = None,
    title_template: str = "Backtesting — Fold {fold_num} (refit: {refit})",
    fps: int = 2,
    dpi: int = 100
) -> str:
    """
    Create a GIF of the backtesting folds using Matplotlib FuncAnimation.

    Parameters
    ----------
    data : pandas Series, pandas DataFrame
        Time series data to be used for backtesting. Can be a single series or 
        multiple series.
    cv : TimeSeriesFold
        TimeSeriesFold object with the information needed to split the data into folds.
    series_to_plot : list, default None
        List of series names to plot. If None, all series will be plotted.
    plot_last_window : bool, default False
        Whether to plot the last window of the time series. If True, window_size 
        must be specified in the cv object. 
    filename : str, default "backtesting.gif"
        Name of the output GIF file.
    figsize : tuple, default (9, 4)
        Size of the figure.
    plt_style : str, default "ggplot"
        Style to use for the plots. See Matplotlib styles for available options 
        here: https://matplotlib.org/stable/gallery/style_sheets/style_sheets_reference.html.
    colors : dict, default None
        Dictionary with colors for the different elements of the plot. Must 
        contain the following keys: ['train', 'last_window', 'gap', 'test', 'v_lines'].
        Defaults colors are {"train": "#329239", "last_window": "#f7931a",
        "gap": "#4d4d4d", "test": "#0d579b", "v_lines": "#252525"}
    title_template : str, default "Backtesting — Fold {fold_num} (refit: {refit})"
        Template for the title of each plot.
    fps : int, default 2
        Frames per second for the GIF.
    dpi : int, default 100
        Dots per inch for the GIF.

    Returns
    -------
    filename_path : str
        Path to the output GIF file.

    """

    if isinstance(data, pd.Series):
        data = input_to_frame(
                   data       = data, 
                   input_name = data.name if data.name is not None else 'y'
               )

    if not isinstance(data, pd.DataFrame):
        raise TypeError(
            f"`data` must be a pandas Series or DataFrame. Got {type(data)}."
        )

    if not isinstance(data.index, (pd.RangeIndex, pd.DatetimeIndex)):
        raise TypeError(
            f"`data` must have a pandas RangeIndex or DatetimeIndex. Got {type(data.index)}."
        )

    cv_name = type(cv).__name__
    if cv_name != "TimeSeriesFold":
        raise TypeError(
            f"`cv` must be a 'TimeSeriesFold' object. Got '{cv_name}'."
        )

    if not isinstance(fps, int) or fps <= 0:
        raise TypeError(f"`fps` must be a positive integer. Got {fps}.")
    if not isinstance(dpi, int) or dpi <= 0:
        raise TypeError(f"`dpi` must be a positive integer. Got {dpi}.")

    if series_to_plot is None:
        series_to_plot = data.columns.to_list()
    else:
        if not isinstance(series_to_plot, list):
            raise TypeError(
                f"`series_to_plot` must be a list of column names. Got {type(series_to_plot)}."
            )
        missing_cols = [col for col in series_to_plot if col not in data.columns]
        if missing_cols:
            raise ValueError(f"Columns not found in `data`: {missing_cols}")

    folds = cv.split(X=data)
    if cv.return_all_indexes:
        # NOTE: +1 to prevent iloc pandas from deleting the last observation
        folds = [
            [
                fold[0],
                [fold[1][0], fold[1][-1] + 1],
                (
                    [fold[2][0], fold[2][-1] + 1]
                    if cv.window_size is not None
                    else []
                ),
                [fold[3][0], fold[3][-1] + 1],
                [fold[4][0], fold[4][-1] + 1],
                fold[5],
            ]
            for fold in folds
        ]

    if colors is None:
        colors = {
            "train": "#329239",
            "last_window": "#f7931a",
            "gap": "#4d4d4d",
            "test": "#0d579b",
            "v_lines": "#252525",
        }
    else:
        if not set(colors.keys()) >= {"train", "last_window", "gap", "test", "v_lines"}:
            raise ValueError(
                f"`colors` must contain the following keys: "
                f"{['train', 'last_window', 'gap', 'test', 'v_lines']}"
            )

    y_values = data[series_to_plot].to_numpy().flatten()
    y_min, y_max = np.nanmin(y_values), np.nanmax(y_values)

    # Include some padding to y-axis limits
    y_pad = 0.05 * (y_max - y_min if np.isfinite(y_max - y_min) and (y_max - y_min) > 0 else 1.0)
    y_min_plot = (y_min - y_pad) if np.isfinite(y_min) else -1
    y_max_plot = (y_max + y_pad) if np.isfinite(y_max) else 1

    with plt.style.context(plt_style):
        fig, ax = plt.subplots(figsize=figsize)

        x_index = data.index

        def _draw_fold(ax, fold, title: str):
            ax.clear()

            for col in series_to_plot:
                ax.plot(
                    x_index,
                    data[col].to_numpy(),
                    linewidth=1.5,
                    alpha=0.9,
                    label=col,
                )

            y_min_relative = (y_min - y_min_plot) / (y_max_plot - y_min_plot)
            y_max_relative = (y_max - y_min_plot) / (y_max_plot - y_min_plot)

            # Train
            train_start, train_end = fold[1]
            ax.axvspan(
                x_index[train_start],
                x_index[train_end],
                y_min_relative,
                y_max_relative,
                facecolor=colors["train"],
                alpha=0.2 if plot_last_window else 0.4,
                zorder=0,
                label="Train",
            )

            # last_window
            if plot_last_window:
                if cv.window_size is None:
                    warnings.warn(
                        "Last window cannot be calculated because the `window_size` "
                        "of the `cv` object is None.",
                        IgnoredArgumentWarning
                    )
                else:
                    last_window_start, last_window_end = fold[2]
                    ax.axvspan(
                        x_index[last_window_start],
                        x_index[last_window_end],
                        y_min_relative,
                        y_max_relative,
                        facecolor=colors["last_window"],
                        alpha=0.4,
                        zorder=0,
                        label="Last window",
                    )

            # Gap (if exists)
            gap_start = fold[3][0]
            gap_end = fold[4][0]
            if gap_start != gap_end:
                ax.axvspan(
                    x_index[gap_start],
                    x_index[gap_end],
                    y_min_relative,
                    y_max_relative,
                    facecolor="#bababa",
                    alpha=0.9,
                    zorder=0,
                    label="Gap",
                    hatch="////",
                    edgecolor=colors["gap"], 
                )

            # Test
            test_start, test_end = fold[4]
            ax.axvspan(
                x_index[test_start],
                x_index[test_end - 1],
                y_min_relative,
                y_max_relative,
                facecolor=colors["test"],
                alpha=0.4,
                zorder=0,
                label="Test",
            )

            def vline(ax, i):
                ax.axvline(
                    x_index[i],
                    ymin=y_min_relative,
                    ymax=y_max_relative,
                    lw=1,
                    ls="--",
                    alpha=0.5,
                    color=colors["v_lines"],
                )

            vline(ax, train_end)
            vline(ax, test_start)

            ax.set_ylim(y_min_plot, y_max_plot)
            ax.set_title(title)
            ax.legend(loc="upper left")
            ax.grid(True, alpha=0.8)

        # --------- Animation functions ----------
        def init():
            # Draw the first fold as the initial state
            _draw_fold(ax, folds[0], title=title_template.format(fold_num=1, refit=0))
            return (ax,)

        def update(i):
            fold = folds[i]
            fold_num = min(i + 1, len(folds) - n_extra)
            refit = fold[5] if len(fold) > 4 else i
            title = title_template.format(fold_num=fold_num, refit=refit)
            _draw_fold(ax, fold, title=title)
            return (ax,)

        # --------- Construction and saving ----------
        # Add extra folds for pause at the end
        n_extra = 2
        for i in range(n_extra):
            folds.append(folds[-1])

        ani = FuncAnimation(
            fig,
            update,
            frames=len(folds),
            init_func=init,
            blit=False, 
            repeat=False
        )

        # Save GIF
        writer = PillowWriter(fps=max(1, int(fps)))
        ani.save(Path(filename).with_suffix('.gif'), writer=writer, dpi=dpi)
        plt.close(fig)

    filename_path = os.path.join(os.getcwd(), filename)

    return Path(filename_path)



# Unit test create_train_X_y ForecasterRecursive
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.recursive import ForecasterRecursive
from sklearn.linear_model import LinearRegression


def test_create_train_X_y_output_when_y_is_series_10_and_exog_is_series_of_float():
    """
    Test the output of create_train_X_y when y=pd.Series(np.arange(10)) and 
    exog is a pandas series of floats.
    """
    y = pd.Series(np.arange(10), dtype=float)
    exog = pd.Series(np.arange(100, 110), name='exog', dtype=float)
    forecaster = ForecasterRecursive(LinearRegression(), lags=5)
    results = forecaster.create_train_X_y(y=y, exog=exog)
    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 105.],
                             [5., 4., 3., 2., 1., 106.],
                             [6., 5., 4., 3., 2., 107.],
                             [7., 6., 5., 4., 3., 108.],
                             [8., 7., 6., 5., 4., 109.]]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 'exog']
        ).astype({'exog': float}),
        pd.Series(
            data  = np.array([5, 6, 7, 8, 9]),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y',
            dtype = float
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_series_equal(results[1], expected[1])
